#include "strops.h"
#include <string.h>
#include <stdlib.h>
#include <assert.h>

#include <stdio.h>

/* Function template
char* strops_(const char* string) {
    char* result = malloc(strlen(string));
    memcpy(result, string, strlen(string));
    return result;
}
*/

char* strops_to_lowercase(const char* string) {
    char* result = malloc(strlen(string));
    memcpy(result, string, strlen(string));
    size_t i;
    for (i = 0; i < strlen(string); i++) {
        if (result[i] >= 'A' && result[i] <= 'Z') {
            result[i] += 32;
        }
    }
    return result;
}

char* strops_to_uppercase(const char* string) {
    char* result = malloc(strlen(string));
    memcpy(result, string, strlen(string));
    size_t i;
    for (i = 0; i < strlen(string); i++) {
        if (result[i] >= 'a' && result[i] <= 'z') {
            result[i] -= 32;
        }
    }
    return result;
}

int strops_is_lowercase(const char* string) {
    size_t i;
    for (i = 0; i < strlen(string); i++) {
        if (!(string[i] >= 'a' && string[i] <= 'z')) {
            return 0;
        }
    }
    return 1;
}

int strops_is_uppercase(const char* string) {
    size_t i;
    for (i = 0; i < strlen(string); i++) {
        if (!(string[i] >= 'A' && string[i] <= 'Z')) {
            return 0;
        }
    }
    return 1;
}

char* strops_insert_at_pos_string(const char* string, const char* string_to_insert, size_t pos) {
    assert(pos <= strlen(string) && "pos needs to be inside string");
    size_t string_length = strlen(string) + strlen(string_to_insert) + 1;
    char* result = malloc(string_length);
    memcpy(result, string, strlen(string));
    size_t i, j;
    /* Make space for string_to_insert */
    for (i = string_length - 1; i > pos; i--) {
        result[i] = result[i - strlen(string_to_insert)];
    }
    /* Insert string into empty space */
    for (j = 0; j < strlen(string_to_insert); j++) {
        result[pos + j] = string_to_insert[j];
    }
    return result;
}

char* strops_remove_at_pos_char(const char* string, size_t pos) {
    assert(pos <= strlen(string) && "pos needs to be inside string");
    char* result = malloc(strlen(string) + 1);
    memcpy(result, string, strlen(string));
    result[pos] = 0;
    size_t i;
    for (i = pos; i < strlen(string); i++) {
        result[i] = result[i + 1];
    }
    return result;
}

char* strops_remove_at_pos_string(const char* string, const char* string_to_remove, size_t pos) {
    char* result = malloc(strlen(string));
    memcpy(result, string, strlen(string));
    return result;
}

char* strops_replace_at_pos_string(const char* string, const char* string_to_remove, const char* string_to_insert, size_t pos) {
    char* result = malloc(strlen(string));
    memcpy(result, string, strlen(string));
    return result;
}

char* strops_trim_right_whitespace(const char* string) {
    char* result = malloc(strlen(string) + 1);
    memcpy(result, string, strlen(string));
    while (strchr("\t\n\v\f\r ", result[strlen(result) - 1]) != NULL) {
        result = strops_remove_at_pos_char(result, strlen(result) - 1);
    }
    result = realloc(result, strlen(result));
    return result;
}

char* strops_trim_left_whitespace(const char* string) {
    char* result = malloc(strlen(string) + 1);
    memcpy(result, string, strlen(string));
    while (strchr("\t\n\v\f\r ", result[0]) != NULL) {
        result = strops_remove_at_pos_char(result, 0);
    }
    result = realloc(result, strlen(result));
    return result;
}

char* strops_trim_both_whitespace(const char* string) {
    char* result = malloc(strlen(string) + 1);
    memcpy(result, string, strlen(string));
    result = strops_trim_right_whitespace(result);
    result = strops_trim_left_whitespace(result);
    return result;
}

char* strops_trim_right_chars(const char* string, const char* chars_to_remove) {
    char* result = malloc(strlen(string) + 1);
    memcpy(result, string, strlen(string));
    while (strchr(chars_to_remove, result[strlen(result) - 1]) != NULL) {
        result = strops_remove_at_pos_char(result, strlen(result) - 1);
    }
    result = realloc(result, strlen(result));
    return result;
}

char* strops_trim_left_chars(const char* string, const char* chars_to_remove) {
    char* result = malloc(strlen(string) + 1);
    memcpy(result, string, strlen(string));
    while (strchr(chars_to_remove, result[0]) != NULL) {
        result = strops_remove_at_pos_char(result, 0);
    }
    result = realloc(result, strlen(result));
    return result;
}

char* strops_trim_both_chars(const char* string, const char* chars_to_remove) {
    char* result = malloc(strlen(string) + 1);
    memcpy(result, string, strlen(string));
    result = strops_trim_right_chars(result, chars_to_remove);
    result = strops_trim_left_chars(result, chars_to_remove);
    return result;
}

char* strops_trim_right_string(const char* string, const char* string_to_remove) {
    assert(strlen(string) >= strlen(string_to_remove) && "string_to_remove cannot be bigger than string");
    char* result = malloc(strlen(string) + 1);
    memcpy(result, string, strlen(string));
    
    size_t offset = strlen(result) - strlen(string_to_remove);
    char* tmp = result + offset;
    while (strcmp(tmp, string_to_remove) == 0) {
        size_t i;
        for (i = 0; i < strlen(string_to_remove); i++) {
            result  = strops_remove_at_pos_char(result, offset);
        }
        offset = strlen(result) - strlen(string_to_remove);
        tmp = result + offset;
    }
    result = realloc(result, strlen(result));
    return result;
}

char* strops_trim_left_string(const char* string, const char* string_to_remove) {
    assert(strlen(string) >= strlen(string_to_remove) && "string_to_remove cannot be bigger than string");
    char* result = malloc(strlen(string) + 1);
    memcpy(result, string, strlen(string));

    char* tmp = malloc(strlen(string_to_remove) + 1);
    memcpy(tmp, result, strlen(string_to_remove));
    while (strcmp(tmp, string_to_remove) == 0) {
        size_t i;
        for (i = 0; i < strlen(string_to_remove); i++) {
            result  = strops_remove_at_pos_char(result, 0);
        }
        memcpy(tmp, result, strlen(string_to_remove));
    }
    free(tmp);
    result = realloc(result, strlen(result));
    return result;
}

char* strops_trim_both_string(const char* string, const char* string_to_remove) {
    assert(strlen(string) >= strlen(string_to_remove) && "string_to_remove cannot be bigger than string");
    char* result = malloc(strlen(string) + 1);
    memcpy(result, string, strlen(string));
    result = strops_trim_right_string(result, string_to_remove);
    result = strops_trim_left_string(result, string_to_remove);
    return result;
}

char* strops_remove_chars(const char* string, const char* chars_to_remove) {
    char* result = malloc(strlen(string) + 1);
    memcpy(result, string, strlen(string));

    size_t i;
    for (i = 0; i < strlen(chars_to_remove); i++) {
        while (strchr(result, chars_to_remove[i]) != NULL) {
            result = strops_remove_at_pos_char(result, strlen(result) - strlen(strchr(result, chars_to_remove[i])));
        }
    }
    result = realloc(result, strlen(result));
    return result;
}

char* strops_remove_string(const char* string, const char* string_to_remove) {
    assert(strlen(string) >= strlen(string_to_remove) && "string_to_remove cannot be bigger than string");
    char* result = malloc(strlen(string) + 1);
    memcpy(result, string, strlen(string));

    while (strstr(result, string_to_remove) != NULL) {
        size_t offset = strlen(result) - strlen(strstr(result, string_to_remove));
        size_t i;
        for (i = 0; i < strlen(string_to_remove); i++) {
            result  = strops_remove_at_pos_char(result, offset);
        }
    }
    result = realloc(result, strlen(result));
    return result;
}

size_t strops_word_count(const char* string) {
    size_t word_count = 0;
    char* tmp = malloc(strlen(string) + 1);
    memcpy(tmp, string, strlen(string));char* strpos_remove_at_pos_string(const char* string, const char* string_to_remove, size_t pos);

    size_t i = 0;
    size_t beginning_of_word = 0;
    size_t end_of_word = end_of_word;
    while (i < strlen(string) - 1) {
        for (; i < strlen(string); i++) {
            if ((tmp[i] >= 'A' && tmp[i] <= 'Z') || (tmp[i] >= 'a' && tmp[i] <= 'z')) {
                beginning_of_word = i;
                break;
            }
        }

        if (strchr("\t\n\v\f\r ", tmp[beginning_of_word - 1]) == NULL) {
            continue;
        }

        for (i = beginning_of_word; i < strlen(string); i++) {
            if (!(tmp[i] >= 'A' && tmp[i] <= 'Z') && !(tmp[i] >= 'a' && tmp[i] <= 'z')) {
                end_of_word = i;
                break;
            }
        }

        if (strchr("\t\n\v\f\r ", tmp[end_of_word + 1]) == NULL) {
            continue;
        }

        if (end_of_word - beginning_of_word < 2) {
            continue;
        }
        
        i = end_of_word;
        word_count++;
    }
    free(tmp);
    return word_count;
}

/* TODO: verify that this is correct */
int strops_is_url_encoded(const char* string) {
    if (strchr(string, '%') != NULL) {
        return 1;
    }
    return 0;
}

char* strops_url_encode(const char* string) {
    if (strops_is_url_encoded(string)) { return NULL; }
    /* multiplied by 3, because example: ( => %28. Three times the chars */
    char* result = malloc(strlen(string) + 1);
    memcpy(result, string, strlen(string));
    char* unsafe_chars = "()";
    char replacement[4];
    replacement[0] = '%';
    replacement[3] = '\0';
    
    size_t i, j;
    for (i = 0; i < strlen(unsafe_chars); i++) {
        j = 0;
        while (strchr(result, unsafe_chars[i]) != NULL) {
            if (result[j] == unsafe_chars[i]) {
                char bad = result[j];
                sprintf(replacement + 1, "%02X", bad);
                result = strops_remove_at_pos_char(result, j);
                result = strops_insert_at_pos_string(result, replacement, j);
                j += 2;
            } else {
                j += 1;
            }
        }
    }
    result = realloc(result, strlen(result));
    return result;
}

char* strops_url_decode(const char* string) {
    if (!strops_is_url_encoded(string)) { return NULL; }
    char* result = malloc(strlen(string) + 1);
    memcpy(result, string, strlen(string));

    result = realloc(result, strlen(result));
    return result;
}

